/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis SOURCE CODE IS (C) COPYRIGHT 1994-2002             *
 * by the XIPHOPHORUS Company http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: simple example decoder

 ********************************************************************/

/* Takes a vorbis bitstream from stdin and writes raw stereo PCM to
   stdout.  Decodes simple and chained OggVorbis files from beginning
   to end.  Vorbisfile.a is somewhat more complex than the code below.  */

/* C#/.NET interop-port
 * 
 * Copyright 2004 Klaus Prckl <klaus.prueckl@aon.at>
 */

using System;
using System.IO;

using Xiph.Interop.Ogg;
using Xiph.Interop.Vorbis;

namespace Xiph.Examples.VorbisDecoder
{
	class Decoder
	{
		private static short [] convbuffer = new short[4096];
		private static int convsize=4096;

		[STAThread]
		static int Main(string[] args)
		{
			OggSyncState	oy = new OggSyncState(); /* sync and verify incoming physical bitstream */
			OggStreamState	os = new OggStreamState(); /* take physical pages, weld into a logical
			  stream of packets */
			OggPage			og = new OggPage(); /* one Ogg bitstream page.  Vorbis packets are inside */
			OggPacket		op = new OggPacket(); /* one raw packet of data for decode */
  
			VorbisInfo		vi = new VorbisInfo(); /* struct that stores all the static vorbis bitstream
			  settings */
			VorbisComment	vc = new VorbisComment(); /* struct that stores all the bitstream user comments */
			// during SynthesisInit vd.vi is set to vi
			VorbisDspState	vd = new VorbisDspState(vi); /* central working state for the packet->PCM decoder */
			VorbisBlock		vb = new VorbisBlock(); /* local working space for packet->PCM decode */
  
			byte[] buffer;
			int  bytes;

#if DEBUG
			BinaryReader stdin = new BinaryReader(new FileStream(@"test.ogg", FileMode.Open, FileAccess.Read));
			BinaryWriter stdout = new BinaryWriter(new FileStream(@"test.raw", FileMode.Create, FileAccess.Write));
#else
			/* We need to set stdin/stdout to binary mode. Damn windows. */
			BinaryReader stdin = new BinaryReader(Console.OpenStandardInput(4096));
			BinaryWriter stdout = new BinaryWriter(Console.OpenStandardOutput());
#endif
			/********** Decode setup ************/

			OggSyncState.Init(oy); /* Now we can read pages */
  
			while(true)
			{ /* we repeat if the bitstream is chained */
				bool eos = false;
				int i;

				/* grab some data at the head of the stream.  We want the first page
				   (which is guaranteed to be small and only contain the Vorbis
				   stream initial header) We need the first page to get the stream
				   serialno. */

				/* submit a 4k block to libvorbis' Ogg layer */
				buffer = OggSyncState.Buffer(oy, 4096);
				bytes=stdin.Read(buffer, 0, 4096);
				OggSyncState.Wrote(oy, bytes);
    
				/* Get the first page. */
				if(OggSyncState.Pageout(oy, og)!=1)
				{
					/* have we simply run out of data?  If so, we're done. */
					if(bytes<4096)break;
      
					/* error case.  Must not be Vorbis data */
					Console.Error.WriteLine("Input does not appear to be an Ogg bitstream.");
					Environment.Exit(1);
				}
  
				/* Get the serial number and set up the rest of decode. */
				/* serialno first; use it to set up a logical stream */
				OggStreamState.Init(os, OggPage.Serialno(og));
    
				/* extract the initial header from the first page and verify that the
				   Ogg bitstream is in fact Vorbis data */
    
				/* I handle the initial header first instead of just having the code
				   read all three Vorbis headers at once because reading the initial
				   header is an easy way to identify a Vorbis bitstream and it's
				   useful to see that functionality seperated out. */
    
				VorbisInfo.Init(vi);
				VorbisComment.Init(vc);
				if(OggStreamState.Pagein(os, og)<0)
				{ 
					/* error; stream version mismatch perhaps */
					Console.Error.WriteLine("Error reading first page of Ogg bitstream data.");
					Environment.Exit(1);
				}
    
				if(OggStreamState.Packetout(os, op)!=1)
				{ 
					/* no page? must not be vorbis */
					Console.Error.WriteLine("Error reading initial header packet.");
					Environment.Exit(1);
				}
    
				if(Vorbis.SynthesisHeaderin(vi, vc, op)<0)
				{ 
					/* error case; not a vorbis header */
					Console.Error.WriteLine("This Ogg bitstream does not contain Vorbis audio data.");
					Environment.Exit(1);
				}
    
				/* At this point, we're sure we're Vorbis.  We've set up the logical
				   (Ogg) bitstream decoder.  Get the comment and codebook headers and
				   set up the Vorbis decoder */
    
				/* The next two packets in order are the comment and codebook headers.
				   They're likely large and may span multiple pages.  Thus we reead
				   and submit data until we get our two pacakets, watching that no
				   pages are missing.  If a page is missing, error out; losing a
				   header page is the only place where missing data is fatal. */
    
				i=0;
				while(i<2)
				{
					while(i<2)
					{
						int result=OggSyncState.Pageout(oy, og);
						if(result==0)break; /* Need more data */
						/* Don't complain about missing or corrupt data yet.  We'll
						   catch it at the packet output phase */
						if(result==1)
						{
							OggStreamState.Pagein(os, og); /* we can ignore any errors here
					 as they'll also become apparent
					 at packetout */
							while(i<2)
							{
								result=OggStreamState.Packetout(os, op);
								if(result==0)break;
								if(result<0)
								{
									/* Uh oh; data at some point was corrupted or missing!
								   We can't tolerate that in a header.  Die. */
									Console.Error.WriteLine("Corrupt secondary header.  Exiting.");
									Environment.Exit(1);
								}
								Vorbis.SynthesisHeaderin(vi, vc, op);
								i++;
							}
						}
					}
					/* no harm in not checking before adding more */
					buffer=OggSyncState.Buffer(oy, 4096);
					bytes = stdin.Read(buffer, 0, 4096);
					if(bytes==0 && i<2)
					{
						Console.Error.WriteLine("End of file before finding all Vorbis headers!");
						Environment.Exit(1);
					}
					OggSyncState.Wrote(oy, bytes);
				}
    
				/* Throw the comments plus a few lines about the bitstream we're
				   decoding */
			{
//				ArrayList[] ptr=vc.user_comments;
//				while(*ptr)
//				{
//					Console.Error.WriteLine("{0}",*ptr);
//					++ptr;
//				}
				Console.Error.WriteLine("\nBitstream is {0} channel, {1}Hz", vi.Channels, vi.Rate);
				Console.Error.WriteLine("Encoded by: {0}\n", vc.Vendor);
			}
    
				convsize=4096/vi.Channels;

				/* OK, got and parsed all three headers. Initialize the Vorbis
				   packet->PCM decoder. */
				Vorbis.SynthesisInit(vd, vi); /* central decode state */
				VorbisBlock.Init(vd, vb);     /* local state for most of the decode
				       so multiple block decodes can
				       proceed in parallel.  We could init
				       multiple VorbisBlock structures
				       for vd here */
    
				/* The rest is just a straight decode loop until end of stream */
				while(!eos)
				{
					while(!eos)
					{
						int result=OggSyncState.Pageout(oy, og);
						if(result==0)break; /* need more data */
						if(result<0)
						{ /* missing or corrupt data at this page position */
							Console.Error.WriteLine("Corrupt or missing data in bitstream; " +
								"continuing...");
						}
						else
						{
							OggStreamState.Pagein(os,og); /* can safely ignore errors at
					 this point */
							while(true)
							{
								result=OggStreamState.Packetout(os, op);

								if(result==0)break; /* need more data */
								if(result<0)
								{ /* missing or corrupt data at this page position */
									/* no reason to complain; already complained above */
								}
								else
								{
									/* we have a packet.  Decode it */
									float[][]pcm;
									int samples;

									///**/                                Console.Error.WriteLine("vb: {0} {1} {2}", vb, vb.vd, vb.reap);
									///**/								Console.Error.WriteLine("op: {0} {1}", op, op.packet.Length);
									
									int ret = 0;
									try 
									{
										ret = Vorbis.Synthesis(vb, op);
									}
									catch(Exception e) 
									{
										Console.Error.WriteLine(e.ToString());
									}
									if(ret==0) /* test for success! */
										Vorbis.SynthesisBlockin(vd, vb);
									/* 
		 
									**pcm is a multichannel float vector.  In stereo, for
									example, pcm[0] is left, and pcm[1] is right.  samples is
									the size of each channel.  Convert the float values
									(-1.<=range<=1.) to whatever PCM format and write it out */

									System.Diagnostics.Debug.Assert(vd.Vi.Channels == 2, "nach Blockin");
									System.Diagnostics.Trace.WriteLine("nach Blockin: " + vd.Vi.Channels);

									while((samples=Vorbis.SynthesisPcmout(vd, out pcm))>0)
									{
										System.Diagnostics.Debug.Assert(vd.Vi.Channels == 2, "nach Pcmout");
										System.Diagnostics.Trace.WriteLine("nach Pcmout: " + vd.Vi.Channels);

										int j;
										bool clipflag = false;
										int bout=(samples<convsize?samples:convsize);
		
										/* convert floats to 16 bit signed ints (host order) and
										   interleave */
										for(i=0;i<vi.Channels;i++)
										{
											int ptr = i;
											int mono = i;
											for(j=0;j<bout;j++)
											{
												//#if true
												int val = (int) (pcm[mono][j]*32767.0f);
												//#else /* optional dither */
												//		    int val=mono[j]*32767.f+drand48()-0.5f;
												//#endif
												/* might as well guard against clipping */
												if(val>32767)
												{
													val=32767;
													clipflag=true;
												}
												if(val<-32768)
												{
													val=-32768;
													clipflag=true;
												}
												convbuffer[ptr] = (short) val;
												ptr += vi.Channels;
											}
										}
		
										if(clipflag)
											Console.Error.WriteLine("Clipping in frame {0}",(int)(vd.Sequence));
		
										for(int k=0; k< vi.Channels*bout; k++)
											stdout.Write(convbuffer[k]);
		
										Vorbis.SynthesisRead(vd, bout); /* tell libvorbis how
						   many samples we
						   actually consumed */
									}	    
								}
							}
							if(OggPage.Eos(og)!=0)
								eos = true;
						}
					}
					if(!eos)
					{
						buffer = OggSyncState.Buffer(oy, 4096);
						bytes = stdin.Read(buffer, 0, 4096);
						OggSyncState.Wrote(oy, bytes);
						if(bytes==0)
							eos = true;
					}
				}
    
				/* clean up this logical bitstream; before exit we see if we're
				   followed by another [chained] */

				OggStreamState.Clear(os);
  
				/* OggPage and OggPacket structs always point to storage in
				   libvorbis.  They're never freed or manipulated directly */
    
				VorbisBlock.Clear(vb);
				VorbisDspState.Clear(vd);
				VorbisComment.Clear(vc);
				VorbisInfo.Clear(vi);  /* must be called last */
			}

			/* OK, clean up the framer */
			OggSyncState.Clear(oy);
  
			GC.KeepAlive(oy);
			GC.KeepAlive(os);
			GC.KeepAlive(og);
			GC.KeepAlive(op);
			GC.KeepAlive(vi);
			GC.KeepAlive(vc);
			GC.KeepAlive(vd);
			GC.KeepAlive(vb);

			Console.Error.WriteLine("Done.");
			return(0);
		}
	}
}
